
#ifndef __CW_VEHICLE_H_
#define __CW_VEHICLE_H_

#include "CarWorldClasses.h"

class CWVehicle;

class Wheel
{
public:
	Wheel(
		CWVehicle *AVehicle,
		const Point3D &RelPos = Point3D(),
		REAL SpringStiffness = 30000,
		REAL MaxTravel = -0.5f,
		REAL WheelRadius = 0.4f,
		REAL DamperCompression = 2000,
		REAL DamperRebound = 2000,
		REAL Grip = 1.0f,
		REAL AccelFactor = 4000,
		REAL BrakeFactor = 4000,
		REAL SteerFactor = .5f,
		bool LockUp = true
	);
	void reset();
	void Set(
		const Ref &ARef,
		HInput *ACommand
	);
	//void Set(Ref &ARef);
	void CalcForce();
	FixedVector &GetRay();
	void draw();
	void writeXMLTag(ostream &out);

//the vehicle this wheel belongs to:
public:
	CWVehicle *m_Vehicle;

//constant for this wheel:
public:
	OFFObject MyOFFObject; //not yet read from .vhc file

	Point3D RelPos;
	REAL SpringStiffness;
	REAL MaxTravel;
	REAL WheelRadius;
	REAL DamperCompression;
	REAL DamperRebound;
	REAL AverageLoad;
	REAL Grip;
	REAL AccelFactor;
	REAL BrakeFactor;
	REAL SteerFactor;
	bool LockUp;

//variables used for computation:
public:
	REAL Thrust;
	REAL Angle;
	bool Lock;
	Point3D Acceleration;
	Ref MyRef;
	Point3D LandSpeed, LandDirection;
	FixedVector Ray; //the ray used to determine the impact with the ground
	REAL Gs;
	REAL Travel, DeltaTravel;
	Contact MyContact;
	Point3D Force;
};

class CWVehicle : public CWFeature
{
public:
	CWVehicle(const char *name, HInput *Input);
	void load(const char *name);
	void save(const char *name);
	virtual ~CWVehicle();
	void reset();
	void update();
	REAL GetSpeed() const; //returns speed in kph

	//update every thing attached to the car
	//ie: roll cage position, wheels, ...
	void UpdateCWVehicleParams(HInput &ACommand);
	void UpdateCWVehicleParams();

	void CalcForces();

	void draw_init();
	//void draw_shutdown();
	void draw();
	void drawShape();
	void ProjectShadow(const Point3D &LightDirection);
	void drawInfo();

	void set_input(HInput *AInput);
public:

//to be replaced by:
	string OffFile;
	Point3D InitPos;
	Point3D ChassisPos;
//mass:
	REAL Mass; //kg
	REAL Weight; //N
	REAL MassDistrib; //kg.m
//to do with drive train:
	//int NB_GEARS 5
	REAL RedLine; //REV.min-1
	REAL MaxTorque; //N.m
	REAL Friction; //N.s2.m-2

	HInput *m_Input;
	DriveTrain MyDriveTrain;
	InertRef MyRef;
	vector<Wheel> Wheels;
	OFFObject MyOFFObject;

};

#endif //__CW_VEHICLE_H_