
#ifndef __H_INPUT_H_
#define __H_INPUT_H_

namespace std {}
using namespace std;

#include "H_Main.h"
#include "H_Standard.h"

#define JOYSTICK_BUTTON_NUMBER 4


class HInput
{
public:
	enum Type {None=0, Key=1, Joy=2};
	//returns AHInputType input if possible or best alternative
	static HInput *GetBestInput(HWindow *w, Type type=HInput::Joy);
public:
	HInput(HWindow *w = NULL);
	virtual void UpdateState();
	virtual const char *GetDescription();
	virtual Type get_type();
	void reset();
	virtual ~HInput();
public:
	REAL Yaxis; //from -1(down) to 1(up)
	REAL Xaxis; //from -1(left) to 1(right)
	bool button[JOYSTICK_BUTTON_NUMBER];

	bool IsValid;
protected:
	HWindow *m_window;
};
ostream &operator<<(ostream &out, const HInput::Type &t);
istream &operator>>(istream &in, HInput::Type &t);

class KeyboardInput : public HInput
{
public:
	KeyboardInput(HWindow *w);
	virtual void UpdateState();
	virtual const char *GetDescription();
	virtual Type get_type();
	virtual ~KeyboardInput();
private:
	H_KEY UpKey, DownKey, LeftKey, RightKey;
	H_KEY ButtonKey[JOYSTICK_BUTTON_NUMBER];
};

#ifdef WIN32

//Windows 95/98 specific code using directX (DirectInput)
#include <dinput.h>

class DXJoystickInput : public HInput
{
public:
	DXJoystickInput(HWindow *m_window);
	virtual void UpdateState();
	virtual const char *GetDescription();
	virtual Type get_type();
	virtual ~DXJoystickInput();

//Windows specific DirectX members and methods
private:
	bool inputInitDirectInput(HINSTANCE hInst, HWND hWnd);
	//this method must be static to be used as a C callback
	//"this" is available through the "This" parameter
	static bool CALLBACK inputEnumDeviceProc(LPDIDEVICEINSTANCE pdidi, DXJoystickInput *This);
	bool inputAcquireDevices();
	bool inputPrepareDevice();
private:
	LPDIRECTINPUT        gpdi;    // base DirectInput object
	LPDIRECTINPUTDEVICE2 gpdiJoystick;    // DirectInputDevice2 objects
	string JoysticProductName;
    GUID guidDevice;
};

#endif //WIN32

#endif //__H_INPUT_H_
