
#ifndef _H_MAIN_H_
#define _H_MAIN_H_

namespace std {}
using namespace std;

#include <vector>
#include <string>
#include <iostream>
#include "H_Keys.h"

#define BH_DEFAULT_X 100
#define BH_DEFAULT_Y 100


//this class is used internally to keep data about the (OS specific) window
class HWindow
{
public:
	//to do with keyboard state: (asynchron)
	virtual bool IsPressed(H_KEY k) = 0;
	virtual const char *GetKeyboardDescription() = 0;
	//make this Window the current opengl rendering window
	virtual void MakeCurrent() = 0;
	virtual void SwapBuffers() = 0;
};

class HAppData;

//derive from this class to create a new (non graphic) application
class HApplication
{
public:
	HApplication();
	virtual ~HApplication();

//precision of the return value of ElapsedTime();
	static int TimeRefreshRate();

	virtual const char *name();
	virtual const char *version();
//elapsed time since last called
	virtual void on_idle(unsigned int elapsed_time);
public:
//OS specific application data, do not access
	HAppData *m_data;
};

//derive from this class to create a new graphic application
//it will render to a (single) opengl window (for now...)
class HglApplication : public HApplication
{
public:
	HglApplication();
	virtual ~HglApplication();

//screen mode manipulation:
	unsigned int GetScreenMode();
	bool SetScreenMode(unsigned int ScreenMode, ostream &out); //SetScreenMode returns true if successful
	void PrintScreenMode(unsigned int ScreenMode, ostream &out); //write screen mode information
	void PrintAllScreenModes(ostream &out);
//initial window parameters:
	virtual unsigned int init_width();
	virtual unsigned int init_height();
	virtual bool init_full_screen();
//events
	virtual void draw_init(); //init is called after the window is opened
	virtual void draw_shutdown(); //init is called befor the window is closed
	virtual void draw();
	virtual void key_down(H_KEY AHKey, char c);
	virtual void resize(unsigned int width, unsigned int height);
public:
//OS specific window
	HWindow *m_window;
};

void Add(HApplication *app);
//void Remove(HApplication *app);
int EventLoop(int argc, char **argv);

#endif //_H_MAIN_H_
