
//#include "StdAfx.h"

#include <math.h>

#include <algorithm>

#include "H_Graphics.h"
#include "H_Object.h"
#include "H_Standard.h"

const Color OFFObject::Polygon::DefaultColor = Grey;

list<HPointer<Texture>::Ref> HPointer<Texture>::RefList;

//CLASS OFFObject::Point:
OFFObject::Vertex::Vertex()
{}
OFFObject::Vertex::Vertex(const Point3D &APosition, const Point3D &ANormal, const Point2D &ATexCoord) :
		Position(APosition),
		Normal(ANormal),
		TexCoord(ATexCoord)
{}
OFFObject::Vertex::Vertex(const OFFObject::Vertex &AVertex) :
		Position(AVertex.Position),
		Normal(AVertex.Normal),
		TexCoord(AVertex.TexCoord)
{}
istream &OFFObject::Vertex::Read(istream &in)
{
	//do not read normal: it is calculated automatically
	in >> Position;
	in >> TexCoord;
	return in;
}
ostream &OFFObject::Vertex::Write(ostream &out) const
{
	out << Position;
	out << TexCoord;
	return out;
}
istream &operator >> (istream &in, OFFObject::Vertex &AVertex)
{
	return AVertex.Read(in);
}
ostream &operator << (ostream &out, const OFFObject::Vertex &AVertex)
{
	return AVertex.Write(out);
}

//CLASS OFFObject::Polygon:
OFFObject::Polygon::Polygon() {}

OFFObject::Polygon::~Polygon() {}

void OFFObject::Polygon::InitNormal(const HVector<OFFObject::Vertex> &OFFVertexes)
{
	Normal = (OFFVertexes[IndexVertexes[2]].Position-OFFVertexes[IndexVertexes[0]].Position)
		^(OFFVertexes[IndexVertexes[1]].Position-OFFVertexes[IndexVertexes[0]].Position);
	Normal.normalize();
}

void OFFObject::Polygon::Invert()
{
	unsigned int swap;
	for (unsigned int k=0 ; k<IndexVertexes.size()/2 ; k++)
	{
		swap = IndexVertexes[k];
		IndexVertexes[k] = IndexVertexes[IndexVertexes.size()-k-1];
		IndexVertexes[IndexVertexes.size()-k-1] = swap;
	}
}

istream &OFFObject::Polygon::Read(istream &in)
{
	unsigned int size;
	in >> size;
	IndexVertexes.resize(size);
	for (unsigned int i=0 ; i<IndexVertexes.size() ; i++)
		in >> IndexVertexes[i];

	//if there is color data, read it
	EatWhite(in);
	if (in.peek() == '0' || in.peek() == '1') // there is a color
		in >> MyColor;
	else //no color found in file
		MyColor = DefaultColor;
	return in;
}
ostream &OFFObject::Polygon::Write(ostream &out) const
{
	out << (IndexVertexes.size());
	for (unsigned int i=0 ; i<IndexVertexes.size() ; i++)
		out << " " << IndexVertexes[i];
	out << "\t" << MyColor;
	return out;
}
istream &operator >> (istream &in, OFFObject::Polygon &APolygon)
{
	return APolygon.Read(in);
}
ostream &operator << (ostream &out, OFFObject::Polygon &APolygon)
{
	return APolygon.Write(out);
}

//CLASS OFFObject::Buffer:
void OFFObject::Buffer::AddVertex(const OFFObject::Vertex &AOFFVertex)
{
	Vertexes.push_back(AOFFVertex);
}

void OFFObject::Buffer::AddPolygon(const OFFObject::Polygon &ATriangle)
{
	Polygons.push_back(ATriangle);
}

void OFFObject::Buffer::SetTexture(const char *ATextureName)
{
	TextureName = ATextureName;
}

void OFFObject::Buffer::ReturnOffObject(OFFObject &tmp)
{
	tmp.MyTexture = (TextureName.empty()) ? NULL : new Texture(TextureName.c_str());
	tmp.Allocate(Vertexes.size(), Polygons.size());

	copy(Vertexes.begin(), Vertexes.end(), tmp.MyOFFVertexes.begin());
	copy(Polygons.begin(), Polygons.end(), tmp.MyOFFPolygons.begin());

	tmp.InitMyData();
}

//CLASS OFFObject:
OFFObject::OFFObject() :
	IsSmooth(true),
	MyTexture(NULL),
	CompiledList(0),
	CompiledList_NoMaterial(0)
{}

void OFFObject::Release()
{
	delete MyTexture;
	MyTexture = NULL;
	HGL_DELETE_LIST(CompiledList);
	HGL_DELETE_LIST(CompiledList_NoMaterial);
}

void OFFObject::Allocate(unsigned int NbVertexs, unsigned int NbPolygons)
{
	MyOFFPolygons.resize(NbPolygons);
	MyOFFVertexes.resize(NbVertexs);
}

bool OFFObject::IsInit()
{
	return MyOFFVertexes.size() != 0;
}

void OFFObject::readfile(const char* FileName, REAL scale)
{
	try
	{
		Release();
		cout << "    " << FileName << "\n";
		ifstream infile;
		OpenForRead(infile, FileName);

		bool IsTextured = false;
		IsSmooth = true;
		if (infile.peek() == 'o' || infile.peek() == 'O')
		{
			char TrashText[150];
			infile.getline(TrashText,150);
			IsTextured = (strcmp(TrashText,"OFF2")==0);
		}
		if (IsTextured)
		{
			char TextureName[150];
			infile.getline(TextureName,150);
			string TextureFile = PathOf(FileName)+'/'+TextureName;
			//H_NEW_ONE(MyTexture, Texture((char*)TextureFile));
			MyTexture = new Texture(TextureFile.c_str());
		}
		unsigned int NVERTICES, NFACES, NEDGES;
		infile >> NVERTICES >> NFACES >> NEDGES;
		Allocate(NVERTICES, NFACES);
		//cout << read all the points from the file
		for (unsigned int i=0 ; i<MyOFFVertexes.size() ; i++)
		{
			infile >> MyOFFVertexes[i].Position;
			if (IsTextured)
				infile >> MyOFFVertexes[i].TexCoord;
		}

		//cout << read all the faces from the file
		for (unsigned int j=0 ; j<MyOFFPolygons.size() ; j++)
			infile >> MyOFFPolygons[j];

		//cout << "initializing OFF data...\n";
		InitMyData(scale);

		//cout << "closing file " << FileName << "\n";
		infile.close();
	}
	catch (const HException &E)
	{
		Release();
		throw HException(string("problem with OFF object \"")+TitleOf(FileName)+"\":\n"+E.what());
	}
}

void OFFObject::writefile(const char* FileName)
{
	ofstream outfile(FileName, ios::out);
	if (MyTexture != 0)
		outfile << "OFF2\n" << (*MyTexture).Title() << "\n";
	else
		outfile << "OFF\n";
	outfile << MyOFFVertexes.size() << " ";
	outfile << MyOFFPolygons.size() << " 0\n";
	//0 is the number of edges, for compatibility with OFF format
	for (unsigned int i=0 ; i<MyOFFVertexes.size() ; i++)
	{
		outfile << MyOFFVertexes[i].Position;
		if (MyTexture != 0)
			outfile << "\t" << MyOFFVertexes[i].TexCoord;
		outfile << endl;
	}
	
	outfile << endl;

	//write all the faces from the file
	for (unsigned int j=0 ; j<MyOFFPolygons.size() ; j++)
		outfile << MyOFFPolygons[j] << endl;

	outfile.close();
}

void OFFObject::InitMyData(REAL scale)
{
	unsigned int i,j;

	HGL_DELETE_LIST(CompiledList);
	HGL_DELETE_LIST(CompiledList_NoMaterial);
	
	if (scale>0)
		CenterAndScale(scale);

	//init the polygon normals
	for (j=0 ; j<MyOFFPolygons.size() ; j++)
		MyOFFPolygons[j].InitNormal(MyOFFVertexes);

	//initialize the vertex normals
	for (j=0 ; j<MyOFFPolygons.size() ; j++)
		for (unsigned int k=0 ; k<MyOFFPolygons[j].IndexVertexes.size() ; k++)
			MyOFFVertexes[MyOFFPolygons[j].IndexVertexes[k]].Normal += MyOFFPolygons[j].Normal;

	for (i=0 ; i<MyOFFVertexes.size() ; i++)
		MyOFFVertexes[i].Normal.normalize();
}

void OFFObject::CenterAndScale(REAL scale)
{
	unsigned int i;
	//calculate the barycenter
	Point3D BaryC(0,0,0);
	for (i=0 ; i<MyOFFVertexes.size() ; i++)
	{
		BaryC += MyOFFVertexes[i].Position;
		MyOFFVertexes[i].Normal.reset();
	}
	BaryC /= (REAL)MyOFFVertexes.size();

	//move the object to the origine
	REAL MaxDist = 0;
	REAL ThisDist;
	for (i=0 ; i<MyOFFVertexes.size() ; i++)
	{
		MyOFFVertexes[i].Position -= BaryC;
		ThisDist = MyOFFVertexes[i].Position.norm();
		if (ThisDist>MaxDist) MaxDist = ThisDist;
	}
	//scale object:
	for (i=0 ; i<MyOFFVertexes.size() ; i++)
		MyOFFVertexes[i].Position = MyOFFVertexes[i].Position*(scale/MaxDist);
}

void OFFObject::GLdraw(bool UseMaterial)
{
	//do we have to build the lists?
	if (!CompiledList) BuildCompiledLists();

	//call the right list
	if (UseMaterial)
		glCallList(CompiledList);
	else
		glCallList(CompiledList_NoMaterial);
}

void OFFObject::GLdraw_NoOpt(bool UseMaterial)
{
	glPushClientAttrib(GL_CLIENT_VERTEX_ARRAY_BIT);

	glEnableClientState(GL_VERTEX_ARRAY);
	glVertexPointer(
		3,
		GL_FLOAT, 
		sizeof(OFFObject::Vertex),
		&(MyOFFVertexes.ptr()->Position)
	);
	if (!UseMaterial)
	{
		//Build the CompiledList_NoMaterial for the shadow
		Hgl::LockArrays(0, MyOFFVertexes.size());
		for (unsigned int i=0 ; i<MyOFFPolygons.size() ; i++)
			glDrawElements(
				GL_POLYGON,           
				MyOFFPolygons[i].IndexVertexes.size(),         
				GL_UNSIGNED_INT,           
				MyOFFPolygons[i].IndexVertexes
			);
	}
	else
	{
		//Build the CompiledList to be called when the object is drawn
		if (IsSmooth)
		{
			glEnableClientState(GL_NORMAL_ARRAY);
			glNormalPointer(GL_FLOAT, sizeof(OFFObject::Vertex), &(MyOFFVertexes.ptr()->Normal));
		}
		if (MyTexture != 0)
		{
			glEnableClientState(GL_TEXTURE_COORD_ARRAY);
			glTexCoordPointer(2, GL_FLOAT, sizeof(OFFObject::Vertex), &(MyOFFVertexes.ptr()->TexCoord));
		}
		Hgl::SetColor(White);
		Hgl::LockArrays(0, MyOFFVertexes.size());
		for (unsigned int i=0 ; i<MyOFFPolygons.size() ; i++)
		{
			if (!IsSmooth) Hgl::Normal(MyOFFPolygons[i].Normal);
			if (MyTexture == 0) Hgl::SetColor(MyOFFPolygons[i].MyColor);
			glDrawElements(
				GL_POLYGON,           
				MyOFFPolygons[i].IndexVertexes.size(),         
				GL_UNSIGNED_INT,           
				MyOFFPolygons[i].IndexVertexes
			);
		}
	}
	Hgl::UnlockArrays();

	glPopClientAttrib();
}

void OFFObject::draw(const Ref &Position)
{
	//is the object initialized?
	if (!IsInit()) return;

	//do we have to use the texture?
	if
	(
		(! Hgl::IsEnabled(Hgl::SHADOW))
		&& (MyTexture != 0)
		&& (Hgl::GetTextureMode() != Hgl::NONE)
	)
		(*MyTexture).Use();
	else
		glDisable(GL_TEXTURE_2D);

	//draw at the correct location
	glPushMatrix();
	Hgl::Relocate(Position);
	//GLdraw_NoOpt(!Hgl::IsEnabled(Hgl::SHADOW));
	GLdraw(!Hgl::IsEnabled(Hgl::SHADOW));
	glPopMatrix();
}

void OFFObject::BuildCompiledLists()
{
	unsigned int i;
	glPushClientAttrib(GL_CLIENT_VERTEX_ARRAY_BIT);

	glEnableClientState(GL_VERTEX_ARRAY);
	glVertexPointer(
		3,
		GL_FLOAT,
		sizeof(OFFObject::Vertex),
		&(MyOFFVertexes.ptr()->Position)
	);
	Hgl::LockArrays(0, MyOFFVertexes.size());
	//Build the CompiledList_NoMaterial for the shadow
	HGL_NEW_LIST(CompiledList_NoMaterial);
	for (i=0 ; i<MyOFFPolygons.size() ; i++)
		glDrawElements(
			GL_POLYGON,           
			MyOFFPolygons[i].IndexVertexes.size(),         
			GL_UNSIGNED_INT,           
			MyOFFPolygons[i].IndexVertexes
		);
	glEndList();

	//Build the CompiledList to be called when the object is drawn
	if (IsSmooth)
	{
		glEnableClientState(GL_NORMAL_ARRAY);
		glNormalPointer(GL_FLOAT, sizeof(OFFObject::Vertex), &(MyOFFVertexes.ptr()->Normal));
	}
	if (MyTexture != 0)
	{
		glEnableClientState(GL_TEXTURE_COORD_ARRAY);
		glTexCoordPointer(2, GL_FLOAT, sizeof(OFFObject::Vertex), &(MyOFFVertexes.ptr()->TexCoord));
	}
	Hgl::SetColor(White); //in case there is a texture
	Hgl::SetColor_optimized();
	HGL_NEW_LIST(CompiledList);
	for (i=0 ; i<MyOFFPolygons.size() ; i++)
	{
		if (!IsSmooth) Hgl::Normal(MyOFFPolygons[i].Normal);
		if (MyTexture == 0) Hgl::SetColor_optimized(MyOFFPolygons[i].MyColor);
		glDrawElements(
			GL_POLYGON,           
			MyOFFPolygons[i].IndexVertexes.size(),         
			GL_UNSIGNED_INT,           
			MyOFFPolygons[i].IndexVertexes
		);
	}
	glEndList();

	Hgl::UnlockArrays();

	glPopClientAttrib();
}

OFFObject::~OFFObject()
{
	Release();
}

REAL OFFObject::GetMaxDist()
{
	REAL MaxDist = 0;
	for (unsigned int i=0 ; i<MyOFFVertexes.size() ; i++)
		MaxDist = max(MaxDist,MyOFFVertexes[i].Position.norm());
	return MaxDist;
}
void OFFObject::InvertNormals()
{
	for (unsigned int i=0 ; i<MyOFFPolygons.size() ; i++)
		MyOFFPolygons[i].Invert();
	InitMyData();
}

void OFFObject::Rotate(const Point3D &K)
{
	for (unsigned int i=0 ; i<MyOFFVertexes.size() ; i++)
		RotateAround(MyOFFVertexes[i].Position,K);
	InitMyData();
}
void OFFObject::Translate(const Point3D &K)
{
	for (unsigned int i=0 ; i<MyOFFVertexes.size() ; i++)
		MyOFFVertexes[i].Position += K;
	InitMyData();
}

void OFFObject::SetShading(bool AIsSmooth)
{
	IsSmooth = AIsSmooth;
	InitMyData();
}

bool OFFObject::GetShading()
{
	return IsSmooth;
}
