
//#include <Xatom.h>

#include <ctype.h>
#include <iostream>
#include "../H_Prompt.h"
#include "../H_Main.h"

/*======================================================================================*/
/*========================X Windows version of the main ================================*/
/*======================================================================================*/

#include <GL/glx.h>
#include <GL/gl.h>

#include <X11/Xlib.h>
#include <X11/Xutil.h>

#include <sys/time.h>

#include <unistd.h> //for sleep()
#include <X11/keysym.h>

#include "../H_Keys.h"

static list<HApplication*> AppList;

H_KEY XKtoHK(int XK)
{
	switch (XK)
	{
	case XK_Cancel : return HK_CANCEL;
	case XK_Clear : return HK_CLEAR;
	case XK_Shift_L : return HK_SHIFT;
	case XK_Shift_R : return HK_SHIFT;
	case XK_Control_L : return HK_CONTROL;
	case XK_Control_R : return HK_CONTROL;
	case XK_Menu : return HK_MENU;
	case XK_Pause : return HK_PAUSE;
	case XK_Escape : return HK_ESCAPE;
	case XK_Prior : return HK_PRIOR;
	case XK_Next : return HK_NEXT;
	case XK_End : return HK_END;
	case XK_Home : return HK_HOME;
	case XK_Select : return HK_SELECT;
	case XK_Print : return HK_PRINT;
	case XK_Execute : return HK_EXECUTE;
	case XK_Insert : return HK_INSERT;
	case XK_Delete : return HK_DELETE;
	case XK_Help : return HK_HELP;
	case XK_KP_0 : return HK_NUMPAD0;
	case XK_KP_1 : return HK_NUMPAD1;
	case XK_KP_2 : return HK_NUMPAD2;
	case XK_KP_3 : return HK_NUMPAD3;
	case XK_KP_4 : return HK_NUMPAD4;
	case XK_KP_5 : return HK_NUMPAD5;
	case XK_KP_6 : return HK_NUMPAD6;
	case XK_KP_7 : return HK_NUMPAD7;
	case XK_KP_8 : return HK_NUMPAD8;
	case XK_KP_9 : return HK_NUMPAD9;
	case XK_multiply : return HK_MULTIPLY;
	case XK_KP_Multiply : return HK_MULTIPLY;
	case XK_KP_Add : return HK_ADD;
	case XK_KP_Separator : return HK_SEPARATOR;
	case XK_KP_Subtract : return HK_SUBTRACT;
	case XK_KP_Decimal : return HK_DECIMAL;
	case XK_KP_Divide : return HK_DIVIDE;
	case XK_Num_Lock : return HK_NUMLOCK;
	case XK_Scroll_Lock : return HK_SCROLL;

	case XK_space : return HK_SPACE;
	case XK_Tab : return HK_TAB;
	case XK_Return : return HK_RETURN;
	case XK_BackSpace : return HK_BACK;
	case XK_Left : return HK_LEFT;
	case XK_Up : return HK_UP;
	case XK_Right : return HK_RIGHT;
	case XK_Down : return HK_DOWN;
	case XK_F1 : return HK_F1;
	case XK_F2 : return HK_F2;
	case XK_F3 : return HK_F3;
	case XK_F4 : return HK_F4;
	case XK_F5 : return HK_F5;
	case XK_F6 : return HK_F6;
	case XK_F7 : return HK_F7;
	case XK_F8 : return HK_F8;
	case XK_F9 : return HK_F9;
	case XK_F10 : return HK_F10;
	case XK_F11 : return HK_F11;
	case XK_F12 : return HK_F12;
	}
	return toupper(XK);
}

int HKtoXK(H_KEY HK)
{
//to be completed...
	switch (HK)
	{
	case HK_SPACE : return XK_space;
	case HK_LEFT : return XK_Left;
	case HK_UP : return XK_Up;
	case HK_RIGHT : return XK_Right;
	case HK_DOWN : return XK_Down;
	}
	return XK_Escape;
}



class HAppData
{
public:
	HAppData();
	int ElapsedTime();
public:
	bool IsInitiated;
	struct timeval MyPreviousTime;
};

HAppData::HAppData() : IsInitiated(false) {}

static int Diff_timeval(struct timeval &A,struct timeval &B)
{
	int tmp = ((A.tv_sec-B.tv_sec)*1000)+(A.tv_usec-B.tv_usec)/1000;
	return tmp;
}

int HAppData::ElapsedTime()
{
	struct timeval MyCurrentTime;
	struct timezone tz;

	if (!IsInitiated)
	{
		gettimeofday(&(MyPreviousTime),&tz);
		IsInitiated = true;
	}
	gettimeofday(&MyCurrentTime,&tz);
	int tmp = Diff_timeval(MyCurrentTime,MyPreviousTime);
	MyPreviousTime = MyCurrentTime;
	return tmp;
}


class HX11Window : public HWindow
{
//overloades
public:
	virtual bool IsPressed(H_KEY k);
	virtual const char *GetKeyboardDescription();
	virtual void MakeCurrent();
	virtual void SwapBuffers();
public:
	HX11Window();
	virtual ~HX11Window();
	bool CheckEvent(XEvent *event);
	void open(const char *name, unsigned int width, unsigned int height, bool full_screen);
public:
	Display *dpy;
	Window w;
	GLXContext cx;
};

HX11Window::HX11Window() : dpy(NULL) {}

bool HX11Window::CheckEvent(XEvent *event)
{
	return XCheckWindowEvent(dpy, w, ~0, event);
}


static Bool WaitForMapNotify(Display *d, XEvent *e, char *arg)
{
	if((e->type == MapNotify) && (e->xmap.window == (Window)arg))
		return GL_TRUE;
	else
		return GL_FALSE;
}

void HX11Window::open(const char *name, unsigned int width, unsigned int height, bool full_screen)
{
	int errorBase,eventBase;
	XVisualInfo           *vi;
	Colormap              cmap;
	XSetWindowAttributes  swa;
	XEvent                event;
	static int attributes[] = {GLX_STENCIL_SIZE, 1, GLX_DEPTH_SIZE, 16, GLX_RGBA, GLX_DOUBLEBUFFER, None};

	dpy = XOpenDisplay(NULL);
	if(dpy == NULL)
		throw HException("could not open display");

	if(!glXQueryExtension(dpy, &errorBase, &eventBase))
		throw HException("could not query extention");

// find an OpenGL-capable Color Index visual with depth buffer
	vi = glXChooseVisual(dpy, DefaultScreen(dpy), attributes);
	if(vi == NULL)
		throw HException("could not get visual");


// create an OpenGL rendering context in DIRECT mode
	cx = glXCreateContext(dpy, vi,  None, GL_TRUE);
	if(cx == NULL)
		throw HException("could not create rendering context");

// create an X colormap since probably not using default visual
	cmap = XCreateColormap(dpy, RootWindow(dpy, vi->screen),
						 vi->visual, AllocNone);

	swa.colormap = cmap;
	swa.border_pixel = 0;
	swa.event_mask = ExposureMask | KeyPressMask | StructureNotifyMask /*| SubstructureNotifyMask
		| ButtonPressMask | ButtonReleaseMask| Button1MotionMask*/;

	w = XCreateWindow(dpy, RootWindow(dpy, vi->screen), 0, 0, width,
		height, 0, vi->depth, InputOutput, vi->visual, CWBorderPixel
		| CWColormap | CWEventMask, &swa);

	XSetStandardProperties(dpy, w, name, name, None, /*argv*/NULL, /*argc*/0, NULL);

	MakeCurrent();

	XMapWindow(dpy, w);
	XIfEvent(dpy, &event, WaitForMapNotify, (char*)w);
//set protocols:
	Atom DelWindowAtom = XInternAtom(dpy,"WM_DELETE_WINDOW",False);
	XSetWMProtocols(dpy, w, &DelWindowAtom, 1);
}

HX11Window::~HX11Window()
{
	XDestroyWindow(dpy,w);
}

void HX11Window::MakeCurrent()
{
	glXMakeCurrent(dpy, w, cx);
}

const char *HX11Window::GetKeyboardDescription()
{
	return "Generic X Windows keyboard";
}

static bool KeysymIsPressed(const char *keys, KeySym AKeySym, Display *dpy)
{
    KeyCode AKeyCode = XKeysymToKeycode(dpy,AKeySym);
    return (keys[AKeyCode/8] & (char)(1 << (AKeyCode%8))) != 0;
}

bool HX11Window::IsPressed(H_KEY k)
{
    char keys_return[32];
    XQueryKeymap(dpy,keys_return);
	return KeysymIsPressed(keys_return, HKtoXK(k),dpy);
}

//swap the screen (or window) buffer
void HX11Window::SwapBuffers()
{
  	glXSwapBuffers(dpy, w);
}


char toASCII(KeySym key)
{
	if (key>=0 && key<=255)
		return char(key);
	else
		return '\0';
}

//default implementations for the application functions:
//CLASS HApplication
HApplication::HApplication() : m_data(new HAppData) {}
HApplication::~HApplication() {delete m_data;}
const char *HApplication::name() {return "generic HApplication";}
const char *HApplication::version() {return "1.0";}
void HApplication::on_idle(unsigned int elapsed_time) {/*sleep(1);*/} //change this...

//CLASS HglApplication
HglApplication::HglApplication() : m_window(new HX11Window()) {}
HglApplication::~HglApplication()
{
	delete m_window;
}
void HglApplication::draw_init() {}
void HglApplication::draw_shutdown() {}
void HglApplication::key_down(H_KEY AHKey, char c) {}
unsigned int HglApplication::init_width() {return 320;}
unsigned int HglApplication::init_height() {return 200;}
bool HglApplication::init_full_screen() {return false;}
void HglApplication::resize(unsigned int width, unsigned int height) {}
void HglApplication::draw() {}


//precision of the return value of ElapsedTime();
int HApplication::TimeRefreshRate()
{
	return 3000;
}


//screen mode manipulation:
unsigned int HglApplication::GetScreenMode()
{
	return 0;
}

bool HglApplication::SetScreenMode(unsigned int ScreenMode, ostream &out)
{
	return  false;
}

void HglApplication::PrintScreenMode(unsigned int ScreenMode, ostream &out) {}

void HglApplication::PrintAllScreenModes(ostream &out) {}

void Add(HApplication *app)
{
	AppList.push_back(app);
	HglApplication *hgl_app = dynamic_cast<HglApplication*>(app);
	if (hgl_app!=NULL)
	{
		dynamic_cast<HX11Window*>(hgl_app->m_window)
			->open(hgl_app->name(), hgl_app->init_width(), hgl_app->init_height(), hgl_app->init_full_screen());
		hgl_app->draw_init();
	}
}

int EventLoop(int argc, char **argv)
{
	while(true)
	{
		//if no more apps, exit
		if (AppList.empty()) return 0;
		//for all apps, process window event or do on_idle
		list<HApplication*>::iterator I=AppList.begin();
		do
		{
			XEvent event;
			HglApplication *app = dynamic_cast<HglApplication *>(*I);
			if (app==NULL)
				(*I)->on_idle((*I)->m_data->ElapsedTime());
			else
			{
				HX11Window * w = dynamic_cast<HX11Window*>(app->m_window);
	 			if (!w->CheckEvent(&event))
	 			{
	 				w->MakeCurrent();
	 				app->on_idle((*I)->m_data->ElapsedTime());
	 			}
	 			else
	 			{
	 				switch(event.type)
	 				{
	 				case KeyPress:
	 					w->MakeCurrent();
	 					KeySym key;
	 					XLookupString((XKeyEvent *)&event, NULL, 0, &key, NULL);
	 					if (XKtoHK(key) == HK_ESCAPE)
	 					{
	 						delete app;
	 						app = NULL;
	 					}
	 					else
	 						app->key_down(XKtoHK(key),toASCII(key));
	 					break;
	 				case ConfigureNotify:
	 					w->MakeCurrent();
	 					app->resize((unsigned int)event.xconfigure.width,(unsigned int)event.xconfigure.height);
	 					break;
	 				case DestroyNotify :
	 					return 0;
	 				case ClientMessage :
	 					app->m_window->MakeCurrent();
	 					delete app;
	 					app = NULL;
	 					break;
	 				}
	 			}
			}
			if (app==NULL)
				I=AppList.erase(I);
			else
				I++;
		} while (I != AppList.end());
	}
}

